//-----------------------------------------------------------------------------
// EXT_TWI_DS1307
// Version 1.5
// 10.11.2011 / UB
// "www.Mikrocontroller-4U.de"
// mc-4u(@)t-online.de
//-----------------------------------------------------------------------------
// Fuer ein externes Modul
// mit dem TWI-IC : DS1307
// EchtzeitUhr (RTC)
// Hinweis : es gibt auch einen Treiber fuer ein PCF8583 RTC-IC
//-----------------------------------------------------------------------------
// das File muss im Ordner
// "myAVR/WinAVR/AVR/Include/UB_EXT_MODULE/"
// liegen und mit #include <UB_EXT_MODULE/EXT_TWI_DS1307.h>
// eingebunden werden
//-----------------------------------------------------------------------------
// Belegt : TWI-BUS
// Benutzt : MK3_2560_TWI.h = Port-D (Bit0 bis Bit1)
//-----------------------------------------------------------------------------
//
// Info zum DS1307
// Grundadresse = 0xD0
// keine Adresspins
//
// Uhrzeitformat : HH:MM:SS [0..23]:[0..59]:[0..59]
// Datumsformat  : DD:MM:YY [1..31].[1..12].[0..99]
//-----------------------------------------------------------------------------

//----------------------------------------------------------------------------
// Includes
//----------------------------------------------------------------------------
#include  "ub_mk3_2560/MK3_2560_TWI.h" //  bindet die TWI-Funktionen ein



//-----------------------------------------------------------------------------
// INFOS zur Benutzung :
// 01. Funktion "EXT_INIT_DS1307" starten
// 02. Funktion "EXT_DS1307_READ_UHR" starten
// 02A. Die Variable : EXT_DS1307_OK zeigt an ob OK oder FEHLER
// 02B. Die Uhr-Daten stehen dann in der Struktur : EXT_DS1307
// 03. Funktion "EXT_DS1307_WRITE_UHR" starten
// 03A. Die Uhr-Daten muessen in der Struktur : EXT_DS1307 stehen
// 03B. Die Variable : EXT_DS1307_OK zeigt an ob OK oder FEHLER
// 04. Funktion "EXT_DS1307_START_UHR" startet die Uhr
// 05. Funktion "EXT_DS1307_STOP_UHR" stoppt die Uhr
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
// Die gelesenen Uhren-Werte werden in einer Struktur uebergeben
// um im externen Programm keine Umwandlung machen zu muessen
// Beispiel 14 Uhr : 23 Minuten : 16 Sekunden
// stunden=14 / minuten=23 / sekunden=16
// beim Datum genauso :
// Beispiel 3. Februar 2001
// tage=3 / monate=2 / jahre=1
//-----------------------------------------------------------------------------
typedef struct {
  char stunden;	        // Anzahl der Stunden  [0 bis 23]
  char minuten;         // Anzahl der Minuten  [0 bis 59]
  char sekunden;		// Anzahl der Sekunden [0 bis 59]
  char tage;            // Anzahl der Tage     [1 bis max 31] je nach Monat
  char monate;          // Anzahl der Monate   [1 bis 12]
  char jahre;           // Anzahl der Jahre    [0 bis 99]
  bool start;           // ist true, wenn die Uhr laueft
}Uhr_t;


//-----------------------------------------------------------------------------
// Funktionen die vom MAIN benutzt werden koennen
//-----------------------------------------------------------------------------
void EXT_INIT_DS1307(void);     // init vom DS1307 (und start der Uhr)
void EXT_DS1307_READ_UHR(void); // auslesen der UHR
void EXT_DS1307_WRITE_UHR(void);// beschreiben der UHR
void EXT_DS1307_START_UHR(void);// starten der UHR
void EXT_DS1307_STOP_UHR(void); // stoppen der UHR


//-----------------------------------------------------------------------------
// Globale Variabeln
//-----------------------------------------------------------------------------
bool EXT_DS1307_OK=false;       // zum ueberpruefen der fehlerfreien Funktion
Uhr_t EXT_DS1307;		// hier stehen die Uhren-Istwerte drinn



//-----------------------------------------------------------------------------
// Register-Definitionen
//-----------------------------------------------------------------------------

#define DS1307_TWI_ADR  0xD0	// Basis-TWI-Adresse vom DS1307 

#define DS1307_REG_0    0x00    // Register0 (nicht aendern)




//-----------------------------------------------------------------------------
// interne Funktionen
//-----------------------------------------------------------------------------
void EXT_DS1307_MAKE_UHR(void);
void EXT_DS1307_MAKE_REG(void);
void EXT_DS1307_RESET(void);
unsigned char EXT_BCD_TO_DEZ(unsigned char bcd,unsigned char maske);
unsigned char EXT_DEZ_TO_BCD(unsigned char dez);

//-----------------------------------------------------------------------------
// interne Globale Variable
//-----------------------------------------------------------------------------
unsigned char EXT_DS1307_REG[7]; // RAM fuer 7 Werte

//-----------------------------------------------------------------------------
void EXT_INIT_DS1307(void) {	// zum initialisieren der externen RTC

  // init vom TWI-BUS
  MK3_INIT_TWI();
  
  // uhr struktur zuruecksetzen
  EXT_DS1307_RESET();

  // jetzt die Uhr automatisch starten
  EXT_DS1307_START_UHR();
}


//-----------------------------------------------------------------------------
void EXT_DS1307_READ_UHR(void) {
  // auslesen des DS1307 und auffuellen der Uhr_Struktur
  // 
  // bei einem Fehler wird EXT_DS1307_OK auf FALSE gesetzt sonst auf TRUE
 
  bool ok=false;
  char antwort=0;
  
  EXT_DS1307_OK=true; // zuerst auf OK

  ok=MK3_TWI_START();	        // sendet Start-Kennung 
  if(ok==true) {
    // nur wenn Startkennung erfolgreich
    antwort=MK3_TWI_SEND(DS1307_TWI_ADR,false); // 1.Byte = Slave-Adresse (ohne ACK)
    if(antwort==TWSR_SLAVE_OK) {
      // Slave hat mit ACK geantwortet
      MK3_TWI_SEND(DS1307_REG_0,false); 	// 2.Byte = Register_0 (ohne ACK)
      MK3_TWI_START();		         		// sendet Start-Kennung
      MK3_TWI_SEND(DS1307_TWI_ADR+1,false);	// 1.Byte = Slave-Adresse+1 (ohne ACK) 
      EXT_DS1307_REG[0]=MK3_TWI_READ(true);     // 2.Byte wird vom Slave gesendet (mit ACK)
      EXT_DS1307_REG[1]=MK3_TWI_READ(true);     // 3.Byte wird vom Slave gesendet (mit ACK)
      EXT_DS1307_REG[2]=MK3_TWI_READ(true);     // 4.Byte wird vom Slave gesendet (mit ACK)
      EXT_DS1307_REG[3]=MK3_TWI_READ(true);     // 5.Byte wird vom Slave gesendet (mit ACK)
      EXT_DS1307_REG[4]=MK3_TWI_READ(true);     // 6.Byte wird vom Slave gesendet (mit ACK)
      EXT_DS1307_REG[5]=MK3_TWI_READ(true);     // 7.Byte wird vom Slave gesendet (mit ACK)
      EXT_DS1307_REG[6]=MK3_TWI_READ(false);    // 8.Byte wird vom Slave gesendet (ohne ACK)
      // auswerten der Register-Werte
      EXT_DS1307_MAKE_UHR();
    }
    else {
      // Slave hat nicht geantwortet 
      EXT_DS1307_OK=false;
    }
    MK3_TWI_STOP();		         // sendet Stop-Kennung 
  }
  else {
    // Startkennung falsch
    EXT_DS1307_OK=false;
  }
  
   
  if(EXT_DS1307_OK==false) {
    // bei einem Fehler wird die Uhr auf 0:00:00 [01.01.0000] gesetzt   
    EXT_DS1307_RESET();   
  }
}


//-----------------------------------------------------------------------------
void EXT_DS1307_WRITE_UHR(void) {
  // beschreiben des DS1307 aus der Uhr_Struktur
  // 
  // bei einem Fehler wird EXT_DS1307_OK auf FALSE gesetzt sonst auf TRUE
  
  bool ok=false;
  char antwort=0;
  
  EXT_DS1307_OK=true; // zuerst auf OK   

  // setzen der Register-Werte
  EXT_DS1307_MAKE_REG();

  ok=MK3_TWI_START();	        // sendet Start-Kennung 
  if(ok==true) {
    // nur wenn Startkennung erfolgreich
    antwort=MK3_TWI_SEND(DS1307_TWI_ADR,false); // 1.Byte = Slave-Adresse (ohne ACK)
    if(antwort==TWSR_SLAVE_OK) {
      // Slave hat mit ACK geantwortet 
      MK3_TWI_SEND(DS1307_REG_0,false);         // 2.Byte = Register_0 (ohne ACK)
      MK3_TWI_SEND(EXT_DS1307_REG[0],false);    // 3.Byte = Wert1 (ohne ACK)
      MK3_TWI_SEND(EXT_DS1307_REG[1],false);    // 4.Byte = Wert2 (ohne ACK)
      MK3_TWI_SEND(EXT_DS1307_REG[2],false);    // 5.Byte = Wert3 (ohne ACK)
      MK3_TWI_SEND(EXT_DS1307_REG[3],false);    // 6.Byte = Wert4 (ohne ACK)
      MK3_TWI_SEND(EXT_DS1307_REG[4],false);    // 7.Byte = Wert5 (ohne ACK)
      MK3_TWI_SEND(EXT_DS1307_REG[5],false);    // 8.Byte = Wert6 (ohne ACK)
      MK3_TWI_SEND(EXT_DS1307_REG[6],false);    // 9.Byte = Wert7 (ohne ACK)
    }
    else {
      // Slave hat nicht geantwortet 
      EXT_DS1307_OK=false;
    }
    MK3_TWI_STOP();		         // sendet Stop-Kennung
  }
  else {
    // Startkennung falsch
    EXT_DS1307_OK=false;
  }
}


//-----------------------------------------------------------------------------
void EXT_DS1307_START_UHR(void) {
  // setzt das Start/Stop-Bit auf Lo
  // damit wird die Uhr gestartet
  //
  // bei einem Fehler wird EXT_DS1307_OK auf FALSE gesetzt sonst auf TRUE

  
  EXT_DS1307_OK=true; // zuerst auf OK

  // zuerst die aktuelle Uhrzeit auslesen
  EXT_DS1307_READ_UHR();

  if((EXT_DS1307_OK==true) && (EXT_DS1307.start==false)){
    // wenn das lesen ok war und die uhr gerade steht
    EXT_DS1307.start=true;  // uhr starten
    EXT_DS1307_WRITE_UHR(); // daten speichern
  }
}



//-----------------------------------------------------------------------------
void EXT_DS1307_STOP_UHR(void) {
  // setzt das Start/Stop-Bit auf Hi
  // damit wird die Uhr angehalten
  //
  // bei einem Fehler wird EXT_DS1307_OK auf FALSE gesetzt sonst auf TRUE


  EXT_DS1307_OK=true; // zuerst auf OK

  // zuerst die aktuelle Uhrzeit auslesen
  EXT_DS1307_READ_UHR();

  if((EXT_DS1307_OK==true) && (EXT_DS1307.start==true)){
    // wenn das lesen ok war und die uhr gerade aktiv ist
    EXT_DS1307.start=false; // uhr stoppen
    EXT_DS1307_WRITE_UHR(); // daten speichern
  }
}



//-----------------------------------------------------------------------------
void EXT_DS1307_MAKE_UHR(void) { // INTERNE-FUNKTION  
  // auswerten der gelesenen Register-Werte
  unsigned char wert=0;

  // sekunden umwandeln
  wert=EXT_BCD_TO_DEZ(EXT_DS1307_REG[0],0x7F);
  if(wert>59) wert=0;
  EXT_DS1307.sekunden=wert;
  // minuten umwandeln
  wert=EXT_BCD_TO_DEZ(EXT_DS1307_REG[1],0x7F);
  if(wert>59) wert=0;
  EXT_DS1307.minuten=wert;
  // stunden umwandeln
  wert=EXT_BCD_TO_DEZ(EXT_DS1307_REG[2],0x3F);
  if(wert>23) wert=0;
  EXT_DS1307.stunden=wert;
  // tage umwandeln
  wert=EXT_BCD_TO_DEZ(EXT_DS1307_REG[4],0x3F);
  if(wert>31) wert=1;
  if(wert<1) wert=1;
  EXT_DS1307.tage=wert;
  // monate umwandeln
  wert=EXT_BCD_TO_DEZ(EXT_DS1307_REG[5],0x1F);
  if(wert>12) wert=1;
  if(wert<1) wert=1;
  EXT_DS1307.monate=wert;
  // jahre umwandeln
  wert=EXT_BCD_TO_DEZ(EXT_DS1307_REG[6],0xFF);
  if(wert>99) wert=0;
  EXT_DS1307.jahre=wert;
  // start/stop auswerten
  wert=EXT_DS1307_REG[0];
  if((wert&0x80)==0) {
    EXT_DS1307.start=true;
  }
  else {
    EXT_DS1307.start=false;
  }
}

//-----------------------------------------------------------------------------
void EXT_DS1307_MAKE_REG(void) {  // INTERNE-FUNKTION  
  // ausfuellen der Register-Werte
  unsigned char wert=0;

  // Werte auf gueltigkeit pruefen
  // bei einem Fehler wird auf minimum gesetzt
  if(EXT_DS1307.stunden>23) EXT_DS1307.stunden=0;
  if(EXT_DS1307.minuten>59) EXT_DS1307.minuten=0;
  if(EXT_DS1307.sekunden>59) EXT_DS1307.sekunden=0;  
  if(EXT_DS1307.tage>31) EXT_DS1307.tage=1;
  if(EXT_DS1307.tage<1) EXT_DS1307.tage=1;
  if(EXT_DS1307.monate>12) EXT_DS1307.monate=1;
  if(EXT_DS1307.monate<1) EXT_DS1307.monate=1;
  if(EXT_DS1307.jahre>99) EXT_DS1307.jahre=0;

  // sekunden umwandeln
  wert=EXT_DEZ_TO_BCD(EXT_DS1307.sekunden);
  if(EXT_DS1307.start==false) wert|=0x80; // Bit7 setzen, uhr=stop
  EXT_DS1307_REG[0]=wert;
  // minuten umwandeln
  wert=EXT_DEZ_TO_BCD(EXT_DS1307.minuten);
  EXT_DS1307_REG[1]=wert;
  // studen umwandeln
  wert=EXT_DEZ_TO_BCD(EXT_DS1307.stunden);
  EXT_DS1307_REG[2]=wert;
  // nc-byte setzen
  wert=EXT_DEZ_TO_BCD(1);
  EXT_DS1307_REG[3]=wert;
  // tage umwandeln
  wert=EXT_DEZ_TO_BCD(EXT_DS1307.tage);
  EXT_DS1307_REG[4]=wert;
  // monate umwandeln
  wert=EXT_DEZ_TO_BCD(EXT_DS1307.monate);
  EXT_DS1307_REG[5]=wert;
  // jahre umwandeln
  wert=EXT_DEZ_TO_BCD(EXT_DS1307.jahre);
  EXT_DS1307_REG[6]=wert;
}


//-----------------------------------------------------------------------------
void EXT_DS1307_RESET(void) {  // INTERNE-FUNKTION  
  // init der Struktur ... auf 0:00:00 Uhr  [Datum 01.01.0000]
  EXT_DS1307.stunden=0;
  EXT_DS1307.minuten=0;
  EXT_DS1307.sekunden=0;
  EXT_DS1307.tage=1;
  EXT_DS1307.monate=1;
  EXT_DS1307.jahre=0;  
}

//-----------------------------------------------------------------------------
unsigned char EXT_BCD_TO_DEZ(unsigned char bcd,unsigned char maske) {  // INTERNE-FUNKTION
  // wandelt BDC-Registerinhalt in Dezimalwert um
  unsigned char ret_wert=0;
  unsigned char lo=0,hi=0; 

  // wert maskieren
  bcd=(bcd&maske);

  // BCD-Wert
  hi=(bcd&0xF0);
  hi=(hi>>4);
  lo=(bcd&0x0F);
  // umrechnen in Dezimal
  ret_wert=(hi*10)+lo;

  return(ret_wert);
}


//-----------------------------------------------------------------------------
unsigned char EXT_DEZ_TO_BCD(unsigned char dez) {  // INTERNE-FUNKTION
  // wandelt Dezimalwert in BCD-Registerinhalt um
  
  unsigned char ret_wert=0;
  unsigned char hi=0,lo=0;
  
  if(dez<10) {
    hi=0x00;
    lo=dez;
  }
  else if((dez>=10) && (dez<20)) {
    hi=0x10;
    lo=dez-10;
  }
  else if((dez>=20) && (dez<30)) {
    hi=0x20;
    lo=dez-20;
  }
  else if((dez>=30) && (dez<40)) {
    hi=0x30;
    lo=dez-30;
  }
  else if((dez>=40) && (dez<50)) {
    hi=0x40;
    lo=dez-40;
  }  
  else if((dez>=50) && (dez<60)) {
    hi=0x50;
    lo=dez-50;
  }  
  else if((dez>=60) && (dez<70)) {
    hi=0x60;
    lo=dez-60;
  }  
  else if((dez>=70) && (dez<80)) {
    hi=0x70;
    lo=dez-70;
  }  
  else if((dez>=80) && (dez<90)) {
    hi=0x80;
    lo=dez-80;
  }  
  else if((dez>=90) && (dez<100)) {
    hi=0x90;
    lo=dez-90;
  }  
  // BCD erstellen  
  ret_wert=(hi|lo);  
  
  return(ret_wert);
}
